## ---------------------------
##
## Script name: ERI algorithm (eri_algorithm_public.R)
##
## Purpose of script: This script is designed to efficiently calculate the Escape Route Index (ERI), a spatially-explicity
##   measure of egress capacity based on terrain and land cover. The background, methods, and demonstrated results can be
##   found in Campbell et al. (currently in review for publication at Fire)
##
## Script Author: Dr. Michael Campbell
##
## Date Created: 2019-06-19
##
## Email: mcampbell@fortlewis.edu
##
## Study Coauthors: Dr. Wesley Page (USFS), Dr. Philip Dennison (U of Utah), Dr. Bret Butler (USFS)
##
## ---------------------------
##
## Notes:
##   Prior to running this script, you will need to acquire the following datasets:
##     - LANDFIRE Existing Vegetation Type (EVT)
##     - USGS 30m Digital Elevation Model (DEM)
##     - A roads and trails dataset (recommendation: FSTopo + Roads and Trails from The National Map combined)
##
##   Depending on the specific travel time modeled, data should be acquired within a significant buffer area around the study area.
##   Once the data are acquired, they all need to be reprojected into the same cartesian coordinate system (ideally, UTM).
##   Generate a slope raster dataset (in degrees) using the DEM.
##   Convert roads and trails into a raster dataset (1 = roads/trails; 0 = background cell values).
##   Ensure that all datasets have the exact same resolution, extent, and cell alignment.
##
##   In addition, the script requires a vector study area boundary.
##
##   Code is designed to work on the following directory and file structure:
##     [base directory]\Input_Data\
##     [base directory]\Input_Data\elevation.tif
##     [base directory]\Input_Data\landfire_evt.tif
##     [base directory]\Input_Data\roads_and_trails.tif
##     [base directory]\Input_Data\slope.tif
##     [base directory]\Input_Data\study_area_boundary.shp
##     [base directory]\Output_Data\
##     [base directory]\Output_Data\Full_Exports\
##
## ---------------------------

# define all of the libraries needed to successfully run the algorithm
list.of.packages <- c("foreach",
                      "doParallel",
                      "raster",
                      "rgdal",
                      "stringr",
                      "gdistance",
                      "rgeos",
                      "concaveman",
                      "foreign")

# create a vector of those packages that are not installed on this computer
new.packages <- list.of.packages[!(list.of.packages %in% installed.packages()[,"Package"])]

# if there are any packages that are not installed, install them
if (length(new.packages) > 0){
  install.packages(new.packages)
}

# load packages
library(foreach)
library(doParallel)
library(parallel)
library(raster)
library(rgdal)
library(stringr)
library(gdistance)
library(rgeos)
library(concaveman)
library(foreign)

# to enhance processing efficiency, you can split this process across multiple computers
# to do so, simply change the parameters below
num.computers <- 1
this.computer <- 1

# print start time
s1 <- Sys.time()
print(s1)

# define number of cores to process data on (by default set to number available on computer)
num.cores <- detectCores()
cluster <- makeCluster(num.cores)
registerDoParallel(cluster)

# set working directory
main.dir <- "C:/Users/mcampbell/Documents/Angeles_NF_Final/Delivery"
setwd(main.dir)

# define other directories
input.dir <- paste0(main.dir, "/Input_Data/")
output.dir <- paste0(main.dir, "/Output_Data/")
full.export.dir <- paste0(output.dir, "Full_Exports/")

# read in input raster datasets
# it is imperative that these inputs have the same resolution and extent
elevation <- raster("Input_Data/elevation.tif")
slope <- raster("Input_Data/slope.tif")
vegetation <- raster("Input_Data/landfire_evt.tif")
roads.trails <- raster("Input_Data/roads_and_trails.tif")

# read in study area boundary
study.area <- readOGR("Input_Data", "study_area_boundary_for_testing")

# reclassify vegetation data from raw vegetation type to WFDSS land cover relative impedance
rat <- read.dbf("Input_Data/landfire_evt.tif.vat.dbf")
colnames(rat)[1] <- "ID"
colnames(rat)[2:ncol(rat)] <- tolower(colnames(rat)[2:ncol(rat)])

g.vals <- rat$ID[rat$evt_order %in% c("No Dominant Lifeform",
                                      "Herbaceous / Nonvascular-dominated",
                                      "Non-vegetated")]
s.vals <- rat$ID[rat$evt_order == "Shrub-dominated"]
t.vals <- rat$ID[rat$evt_order == "Tree-dominated"]
w.vals <- rat$ID[rat$classname == "Open Water"]
g.vals <- g.vals[g.vals != w.vals]

rc <- as.matrix(data.frame(from = c(g.vals,s.vals,t.vals,w.vals),
                           to = c(rep(1,length(g.vals)),
                                  rep(0.5,length(s.vals)),
                                  rep(0.25,length(t.vals)),
                                  rep(1.0/300,length(w.vals)))))

vegetation <- reclassify(vegetation,rc)

# burn roads and trails into vegetation
vegetation <- overlay(vegetation,roads.trails,fun = function(x,y){
  x[y[] == 1] <- 1
  return(x)
})

# reclassify slope data (1 = passable; 0 = impassable)
slope[slope < 45] <- 1
slope[slope >= 45] <- 0

# define output coordinate system
coord.sys <- crs(elevation)

# define output cell size
cell.size.x <- res(elevation)[1]
cell.size.y <- res(elevation)[2]

# define slope travel rate function terms, based on the 5th percentile Lorentz function from Campbell et al. (2019)
a = -1.5265
b = 14.0405
c = 36.8130
d = 0.3200
e = -0.0027

# create slope travel rate function using the terms above
slope.travel.rate.fun <- function(slope){
  (c*(1/(pi*b*(1+((slope-a)/b)^2)))+d)+e*slope
}

# optimize slope travel rate function to find maximum pace
o <- optimize(slope.travel.rate.fun,c(-30,30),maximum=T)
max.pace <- 1/o$objective

# define the travel time (in seconds) for which ERI will be calculated
travel.time <- 1800
travel.dist <- travel.time / max.pace
travel.dist.round <- cell.size.x * round(travel.dist / cell.size.x)

# calculate the travel distance octagon's circuradius, inradius, and average radius based on travel time
circum.radius <- travel.dist
in.radius <- circum.radius * 0.9238795325
average.radius <- (circum.radius + in.radius) / 2

# clip elevation data to study area boundary
elev.clip <- mask(elevation, study.area)
elev.clip <- crop(elev.clip, study.area)

# determine how to split input data up into n chunks, where n is the number of computers being used
if (num.computers > 1){
  
  # create a data frame that will cumulatively add up non-NA cell counts
  split.df <- data.frame(row = seq(1,nrow(elev.clip)),
                         cell.count = rowSums(!is.na(elev.clip)))
  split.df$run.count <- cumsum(split.df$cell.count)
  
  # calculate the total number of non-NA cells and split it into n chunks, where n is the number of computers being used
  total.count <- split.df$run.count[dim(split.df)[1]]
  even.split.count <- total.count/num.computers
  
  # create a data frame that will determine the starting and ending rows of n even chunks
  even.split.df <- data.frame(chunk = seq(1,num.computers),
                              start.row = rep(NA,num.computers),
                              end.row = rep(NA,num.computers))
  
  # loop through each computer and figure out the starting and ending rows
  for (i in seq(1,num.computers)){
    
    # if this is the first computer...
    if (i == 1){
      
      # the first starting row will be 1, the first ending row will be the closest row number to 1 + the even split count
      even.split.df$start.row[i] <- 1
      even.split.df$end.row[i] <- split.df$row[which.min(abs(even.split.count * i - split.df$run.count))]
      
      # the second starting row will be one greater than the first ending row
      even.split.df$start.row[i+1] <- even.split.df$end.row[i]+1
      
      # if this is the last computer...
    } else if (i == num.computers){
      
      # the last ending row will be the total number of rows
      even.split.df$end.row[i] <- nrow(elev.clip)
      
      # if it's any computer in between...
    } else {
      
      # the ending row will be the closest row number to the starting row + the even split count
      # and the next starting row will be one greater than the ending row
      even.split.df$end.row[i] <- split.df$row[which.min(abs(even.split.count * i - split.df$run.count))]
      even.split.df$start.row[i+1] <- even.split.df$end.row[i]+1
      
    }
    
  }
  
} else { # if it's the only computer..
  
  # create a data frame with just one chunk, where the starting row is 1, and the ending row is the total number of rows
  even.split.df <- data.frame(chunk = 1,
                              start.row = 1,
                              end.row = nrow(elev.clip))
  
}

# get raster chunk for this computer, crop the elevation dataset to that chunk of rows, and define a new extent
start.row <- even.split.df$start.row[this.computer]
end.row <- even.split.df$end.row[this.computer]
elev.clip <- crop(elev.clip, extent(elev.clip, start.row, end.row, 1, ncol(elev.clip)))
new.extent <- extent(elev.clip)

# define cell center coordinate ranges in x and y for the raster chunk extent
range.y <- seq(new.extent[4] - cell.size.y / 2, new.extent[3] + cell.size.y / 2, -cell.size.y)
range.x <- seq(new.extent[1] + cell.size.x / 2, new.extent[2] - cell.size.x / 2, cell.size.x)

# define number of rows and columns for output raster dataset
rows <- length(range.y)
cols <- length(range.x)

# get rows and columns order of magnitude
rows.order <- nchar(as.character(rows))
cols.order <- nchar(as.character(cols))

# if the computer has multiple cores, figure out starting and ending rows for parallel processing
# (same process as splitting up between computers, but now within a single computer raster chunk)
if (num.cores > 1){
  
  # create a data frame that will cumulatively add up non-NA cell counts
  split.df <- data.frame(row = seq(1,nrow(elev.clip)),
                         cell.count = rowSums(!is.na(elev.clip)))
  split.df$run.count <- cumsum(split.df$cell.count)
  
  # calculate the total number of non-NA cells and split it into n chunks, where n is the number of cores being used
  total.count <- split.df$run.count[dim(split.df)[1]]
  even.split.count <- total.count/num.cores
  
  # create a data frame that will determine the starting and ending rows of n even chunks
  even.split.df <- data.frame(chunk = seq(1,num.cores),
                              start.row = rep(NA,num.cores),
                              end.row = rep(NA,num.cores))
  
  # loop through each core and figure out the starting and ending rows
  for (i in seq(1,num.cores)){
    
    # if this is the first core...
    if (i == 1){
      
      # the first starting row will be 1, the first ending row will be the closest row number to 1 + the even split count
      even.split.df$start.row[i] <- 1
      even.split.df$end.row[i] <- split.df$row[which.min(abs(even.split.count * i - split.df$run.count))]
      
      # the second starting row will be one greater than the first ending row
      even.split.df$start.row[i+1] <- even.split.df$end.row[i]+1
      
      # if this is the last core...
    } else if (i == num.cores){
      
      # the last ending row will be the total number of rows
      even.split.df$end.row[i] <- nrow(elev.clip)
      
      # if it's any core in between...
    } else {
      
      # the ending row will be the closest row number to the starting row + the even split count
      # and the next starting row will be one greater than the ending row
      even.split.df$end.row[i] <- split.df$row[which.min(abs(even.split.count * i - split.df$run.count))]
      even.split.df$start.row[i+1] <- even.split.df$end.row[i]+1
      
    }
    
  }
  
} else { # if it's the only core...
  
  # create a data frame with just one chunk, where the starting row is 1, and the ending row is the total number of rows
  even.split.df <- data.frame(chunk = 1,
                              start.row = 1,
                              end.row = nrow(elev.clip))
  
}

# figure out starting rows for parallel processing
starts <- even.split.df$start.row
ends <- even.split.df$end.row
original.starts <- starts

# check if there are already outputs generated; if so, determine where they left off and start there
restart.flag <- 0
interim.outputs <- list.files(output.dir, "*.tif")
if (length(interim.outputs) > 0){
  new.starts <- c()
  for (start in starts){
    interim.outputs.start <- interim.outputs[grep(paste0("row_",str_pad(start,rows.order,"left","0")),interim.outputs)]
    new.start <- str_split(interim.outputs.start[1],"_")
    new.start <- new.start[[1]][6]
    new.start <- as.integer(str_sub(new.start,1,-4))
    new.starts <- append(new.starts,new.start)
  }
  starts <- new.starts
  restart.flag <- 1
}

# create function
eri.fun <- function(n){
  
  # define starting and ending row
  original.start <- original.starts[n]
  start <- starts[n]
  end <- ends[n]
  
  # create a new extent for the chunk
  chunk.extent <- extent(c(new.extent[1],
                           new.extent[2],
                           range.y[end] - cell.size.y / 2,
                           range.y[start] + cell.size.y / 2))
  
  # conditional to determine if this is a brand new start, or continuation from previous distruption
  if (restart.flag == 0){
    
    # create results rasters to write to for ERImax, ERImin, ERImean, and ERIazimuth
    results.raster.max <- raster(chunk.extent, nrows = (end - start) + 1, ncols = cols, crs = coord.sys)
    results.raster.min <- raster(chunk.extent, nrows = (end - start) + 1, ncols = cols, crs = coord.sys)
    results.raster.mean <- raster(chunk.extent, nrows = (end - start) + 1, ncols = cols, crs = coord.sys)
    results.raster.azimuth <- raster(chunk.extent, nrows = (end - start) + 1, ncols = cols, crs = coord.sys)

  } else {
    
    # define original starts
    original.start <- original.starts[n]
    
    # convert original start and start values to padded string for reading in rasters
    original.start.str <- str_pad(original.start,rows.order,"left","0")
    start.str <- str_pad(start,rows.order,"left","0")
    
    # read in existing results rasters to write to for ERImax, ERImin, ERImean, and ERIazimuth
    results.raster.max <- raster(paste0("Output_Data/eri_max_row_", original.start.str, "_", start.str, ".tif"))
    results.raster.min <- raster(paste0("Output_Data/eri_min_row_", original.start.str, "_", start.str, ".tif"))
    results.raster.mean <- raster(paste0("Output_Data/eri_mean_row_", original.start.str, "_", start.str, ".tif"))
    results.raster.azimuth <- raster(paste0("Output_Data/eri_azimuth_row_", original.start.str, "_", start.str, ".tif"))

  }
  
  # create result raster vectors
  results.rasters <- c(results.raster.max,
                       results.raster.min,
                       results.raster.mean,
                       results.raster.azimuth)
  
  # set -0.999 to be the nodata values for all rasters
  for (r in results.rasters){
    NAvalue(r) <- -0.999
  }
  
  # loop through all rows (y)
  for (row in seq(start,end)){
    
    # calculate the results raster row
    raster.row <- row - original.start + 1
    
    # get non-na cols in this row
    data.cols <- which(!is.na(elev.clip[row,]))
    if (length(data.cols) == 0){
      next
    }
    
    # get a random column number for exporting
    rand.col <- sample(data.cols,1)

    # fill in the na cells with -0.999
    nodata.cols <- which(is.na(elev.clip[row,]))
    for (r in results.rasters){
      r[raster.row,nodata.cols] <- -0.999
    }
    
    # loop through all columns (x)
    for (col in data.cols){
      
      # define the coordinates for the starting point and create spatial points
      x0 <- range.x[col]
      y0 <- range.y[row]
      starting.pt <- SpatialPoints(data.frame(x = x0, y = y0), coord.sys)
      
      # get subset coordinates
      subset.xmin <- x0 - (travel.dist.round + cell.size.x / 2)
      subset.xmax <- x0 + (travel.dist.round + cell.size.x / 2)
      subset.ymin <- y0 - (travel.dist.round + cell.size.y / 2)
      subset.ymax <- y0 + (travel.dist.round + cell.size.y / 2)
      
      # subset the elevation dataset
      subset.elev <- crop(elevation,extent(c(subset.xmin,
                                             subset.xmax,
                                             subset.ymin,
                                             subset.ymax)))
      
      # subset the slope dataset
      subset.slope <- crop(slope,extent(c(subset.xmin,
                                          subset.xmax,
                                          subset.ymin,
                                          subset.ymax)))
      
      # subset the vegetation dataset
      subset.veg <- crop(vegetation,extent(c(subset.xmin,
                                             subset.xmax,
                                             subset.ymin,
                                             subset.ymax)))
      
      # check if the sarting point falls on a slope restricted area
      if (extract(subset.slope,starting.pt) == 0){
        for (r in results.rasters){
          r[raster.row,col] <- 0
        }
        next
      }
      
      # calculate slope-based conductance
      slope.conductance <- transition(subset.slope, min, 8, symm=T)
      
      # calculate vegetation-based conductance
      veg.conductance <- transition(subset.veg, mean, 8, symm=T)
      
      # calculate elevation-based conductance
      subset.elev.diff <- transition(subset.elev, function(x){x[2]-x[1]}, 8, symm=F)
      subset.slope.percent <- geoCorrection(subset.elev.diff)
      subset.adjacency <- adjacent(subset.elev, cells = 1:ncell(subset.elev), pairs = T, directions = 8)
      subset.travel.rate <- subset.slope.percent
      subset.travel.rate[subset.adjacency] <- (c*(1/(pi*b*(1+(((subset.slope.percent[subset.adjacency]*180/pi)-a)/b)^2)))+d)+e*(subset.slope.percent[subset.adjacency]*180/pi)
      elev.conductance <- geoCorrection(subset.travel.rate)
      
      # modify vegetation-based conductance on steep slopes for timber to timber
      veg.conductance[subset.adjacency][which(veg.conductance[subset.adjacency] == 0.25 &
                                                abs(subset.slope.percent[subset.adjacency]) > 10 * pi / 180)] <- 1/6
      
      # modify vegetation-based conductance on steep slopes for timber to brush
      veg.conductance[subset.adjacency][which(veg.conductance[subset.adjacency] == 0.375 &
                                                abs(subset.slope.percent[subset.adjacency]) > 10 * pi / 180)] <- (1/6+1/2)/2
      
      # modify vegetation-based conductance on steep slopes for timber to grass
      veg.conductance[subset.adjacency][which(veg.conductance[subset.adjacency] == 0.625 &
                                                abs(subset.slope.percent[subset.adjacency]) > 10 * pi / 180)] <- (1/6+1)/2
      
      
      # calculate overall conductance
      total.conductance <- elev.conductance * slope.conductance * veg.conductance
      
      # calculate travel time
      possible.error <- tryCatch(
        travel.time.rast <- accCost(total.conductance, c(x0,y0)),
        error = function(e) e
      )
      
      # handle errors
      if(inherits(possible.error,"error")){
        for (r in results.rasters){
          r[raster.row,col] <- -0.999
        }
        next
      }
      
      # apply travel time threshold
      travel.time.rast[travel.time.rast > travel.time] <- NA
      
      # convert to points
      possible.error <- tryCatch(
        {travel.time.pts <- as(travel.time.rast, "SpatialPoints")},
        error = function(e) e
      )
      
      # handle errors
      if(inherits(possible.error,"error")){
        for (r in results.rasters){
          r[raster.row,col] <- 0
        }
        next
      }
      
      # generate concave hull, sample points around the edge, and calculate distances to points from starting point
      travel.time.poly <- concaveman(travel.time.pts, 1, 2 * cell.size.x)
      travel.time.points <- spsample(as(travel.time.poly, "SpatialLines"), 100, "regular")
      point.distances <- pointDistance(c(x0,y0),travel.time.points, lonlat=F)

      # calculate eri max, min, and mean values and apply to raster cells
      eri.max <- max(point.distances) / travel.dist
      eri.min <- min(point.distances) / travel.dist
      eri.mean <- mean(point.distances) / average.radius

      results.raster.max[raster.row,col] <- eri.max
      results.raster.min[raster.row,col] <- eri.min
      results.raster.mean[raster.row,col] <- eri.mean

      # calculate eri azimuth and apply to raster cells
      eri.azimuth <- atan2(coordinates(travel.time.points)[which.max(point.distances),2] - y0,
                              coordinates(travel.time.points)[which.max(point.distances),1] - x0)
      eri.azimuth <- eri.azimuth * (180 / pi)
      eri.azimuth <- (360 - eri.azimuth + 90) %% 360

      results.raster.azimuth[raster.row,col] <- eri.azimuth

      # if the column number is the random column number, export a bunch of data (for visualization/testing)
      if (col == rand.col){
        
        # create padded row and column strings for exporting folder names
        row.str <- str_pad(row,rows.order,"left","0")
        col.str <- str_pad(col,cols.order,"left","0")
        
        # create the row-column export directory
        export.dir.rc <- paste0(full.export.dir, "Row_", row.str, "_Col_", col.str)
        dir.create(export.dir.rc, showWarnings = F)
        
        # set the working directory to the row-column export directory
        setwd(export.dir.rc)
        
        # convert spatial features to data frames
        starting.pt.spdf <- SpatialPointsDataFrame(starting.pt, data.frame(id = 1), proj4string = coord.sys)
        furthest.pt.spdf <- SpatialPointsDataFrame(travel.time.points[which.max(point.distances)], data.frame(id = 1), proj4string = coord.sys)
        travel.time.points.spdf <- SpatialPointsDataFrame(travel.time.points, data.frame(id = rep(1,length(travel.time.points))), proj4string = coord.sys)
        travel.time.poly.spdf <- SpatialPolygonsDataFrame(travel.time.poly, data.frame(id = rep(1,length(travel.time.poly))), match.ID = F)

        # export them to shapefile
        writeOGR(starting.pt.spdf, ".", "starting_point", driver = "ESRI Shapefile", overwrite_layer = T)
        writeOGR(furthest.pt.spdf, ".", "furthest_point", driver = "ESRI Shapefile", overwrite_layer = T)
        writeOGR(travel.time.points.spdf, ".", "edge_points", driver = "ESRI Shapefile", overwrite_layer = T)
        writeOGR(travel.time.poly.spdf, ".", "travel_distance", driver = "ESRI Shapefile", overwrite_layer = T)

        # write the travel time raster to file
        writeRaster(travel.time.rast, "travel_time.tif", overwrite = T)
        
        # reset the working directory
        setwd(main.dir)
        
      }
      
    }
    
    # create padded strings for exporting rasters
    original.start.str <- str_pad(original.start, rows.order, "left", "0")
    row.str <- str_pad(row,rows.order, "left", "0")
    row.prev.str <- str_pad(row-1, rows.order, "left", "0")
    
    # after each row in the chunk, write temporary output raster datasets
    writeRaster(results.raster.max, filename = paste0("Output_Data/eri_max_row_", original.start.str, "_", row.str, ".tif"), overwrite = T)
    writeRaster(results.raster.min, filename = paste0("Output_Data/eri_min_row_", original.start.str, "_", row.str, ".tif"), overwrite = T)
    writeRaster(results.raster.mean, filename = paste0("Output_Data/eri_mean_row_", original.start.str, "_", row.str, ".tif"), overwrite = T)
    writeRaster(results.raster.azimuth, filename = paste0("Output_Data/eri_azimuth_row_", original.start.str, "_", row.str, ".tif"), overwrite = T)

    # delete the previous temporary files
    file.remove(paste0("Output_Data/eri_max_row_", original.start.str, "_", row.prev.str, ".tif"))
    file.remove(paste0("Output_Data/eri_min_row_", original.start.str, "_", row.prev.str, ".tif"))
    file.remove(paste0("Output_Data/eri_mean_row_", original.start.str, "_", row.prev.str, ".tif"))
    file.remove(paste0("Output_Data/eri_azimuth_row_", original.start.str, "_", row.prev.str, ".tif"))

  }
  
  # after all of the rows have been completed, write final chunk rasters
  writeRaster(results.raster.max, filename = paste0("Output_Data/eri_max_final_chunk_", n, ".tif"), overwrite = T)
  writeRaster(results.raster.min, filename = paste0("Output_Data/eri_min_final_chunk_", n, ".tif"), overwrite = T)
  writeRaster(results.raster.mean, filename = paste0("Output_Data/eri_mean_final_chunk_", n, ".tif"), overwrite = T)
  writeRaster(results.raster.azimuth, filename = paste0("Output_Data/eri_azimuth_final_chunk_", n, ".tif"), overwrite = T)

  # delete the previous temporary files
  file.remove(paste0("Output_Data/eri_max_row_", original.start.str, "_", row.str,".tif"))
  file.remove(paste0("Output_Data/eri_min_row_", original.start.str, "_", row.str,".tif"))
  file.remove(paste0("Output_Data/eri_mean_row_", original.start.str, "_", row.str,".tif"))
  file.remove(paste0("Output_Data/eri_azimuth_row_", original.start.str, "_", row.str,".tif"))

}

# run the process in parallel
foreach(n=seq(1,length(starts)),
        .verbose = T,
        .packages = c("raster",
                      "rgdal",
                      "stringr",
                      "gdistance",
                      "rgeos",
                      "concaveman")) %dopar% {eri.fun(n)}

# close threads
stopCluster(cluster)

# # run the process on a single thread
# eri.fun(1)

# print end time
s2 <- Sys.time()
print(s2)

# print elapsed time
elapsed <- s2 - s1
print(elapsed)
